
import java.lang.ref.*;
import java.util.*;

public class RefTest {

final static class Mutex {

  private boolean locked = false;

  public void lock() {
    locked = true;
  }

  public void unlock() {
    locked = false;
  }

  public boolean isLocked() {
    return locked;
  }
}

static class MutexFactory {
  public static Mutex newMutex() {
    Mutex mutex = new Mutex();
    MutexRef ref =  
      new MutexRef( mutex, Singleton.getSingleton().getQueue() );
    Singleton.getSingleton().getRefs().add( ref );
    return mutex;
  }
}

static class MutexRef extends WeakReference {

  private Throwable trace;

  public MutexRef( Mutex mutex ) {
    super( mutex );
    trace = new Throwable();
  }

  public MutexRef( Mutex mutex, ReferenceQueue queue ) {
    super( mutex, queue );
    trace = new Throwable();
  }

  public void clear() {
    System.out.println( "Being cleared." );
    Mutex mutex = ( Mutex ) get();
    if ( mutex.isLocked() ) {
      String msg = "The client forgot to unlock the mutex." +
                   "The mutex was created at: ";
      System.out.println( msg );
      trace.printStackTrace();
    }
    super.clear();
  }
}

static class Singleton {
  private static Singleton singleton = new Singleton();
  private List refs = new ArrayList();
  private ReferenceQueue queue = new ReferenceQueue();

  public static Singleton getSingleton() { return singleton; }
  public List getRefs() { return refs; }

  public ReferenceQueue getQueue() { return queue; }
}

public static void main( String[] args ) throws Exception {

  loseMutex();

  /* Generate some garbage so the collector will run.
   */
  for ( int i = 0; i < 100000; ++i ) {
    new Object();
  }

  /* Get the MutexRef off of the queue.
   */
  MutexRef ref = ( MutexRef ) 
    Singleton.getSingleton().getQueue().poll();
  System.out.println( "Mutex reference: " + ref.get() );
}

public static void loseMutex() {
  Mutex m = MutexFactory.newMutex();
  m.lock();
}

}
